/*
 * Simple XQueryExtension function filter.
 *
 * Copyright (C) 2005-2019  Dmitry V. Levin <ldv@altlinux.org>
 * All rights reserved.
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 *
 * gcc -shared -fPIC -Wall -W -Werror filterXQueryExtension.c -ldl -o filterXQueryExtension.so
 */

#ifndef _GNU_SOURCE
# define _GNU_SOURCE
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <dlfcn.h>

static const char *allowed_list;
static int verbose;

static void __attribute__((constructor))
init(void)
{
	allowed_list = getenv("XQueryExtensionAllowed");
	if (allowed_list)
		allowed_list = strdup(allowed_list);
	if (!allowed_list)
		allowed_list = "";

	const char *p = getenv("XQueryExtensionVerbose");
	verbose = p && atoi(p);
}

static void *
load_sym(const char *name)
{
	(void) dlerror();
	void *addr = dlsym(RTLD_NEXT, name);
	const char *msg = dlerror();
	if (msg) {
		fprintf(stderr, "dlsym(%s): %s\n", name, msg);
		_exit(1);
	}
	return addr;
}

static int
is_allowed(const char *name)
{
	int allowed = 0;
	int not = allowed_list[0] == '!' ? 1 : 0;
	const char *p = allowed_list + not;

	while (*p) {
		size_t len = strcspn(p, ",");
		if (len) {
			allowed = strncmp(name, p, len) == 0 && name[len] == '\0';
			if (allowed)
				break;
			p += len;
		}
		p += strspn(p, ",");
	}

	return not ? !allowed : allowed;
}

int
XQueryExtension(void *dpy, const char *name, int *a1, int *a2, int *a3)
{
	static const char fname[] = "XQueryExtension";

	if (is_allowed(name)) {
		static int (*next)(void*, const char*, int*, int*, int*);
		if (!next)
			next = load_sym(fname);

		int rc = next(dpy, name, a1, a2, a3);
		if (verbose)
			fprintf(stderr,
				"### %s(%p, %s, [%d], [%d], [%d]) = %d\n",
				fname, dpy, name, *a1, *a2, *a3, rc);
		return rc;
	} else {
		if (verbose)
			fprintf(stderr,
				"### %s(%p, %s, %p, %p, %p) = 0 (FILTERED)\n",
				fname, dpy, name, a1, a2, a3);
		return 0;
	}
}

char **
XListExtensions(void *dpy, int *nextensions)
{
	static const char fname[] = "XListExtensions";
	static char ** (*next)(void*, int*);
	if (!next)
		next = load_sym(fname);

	char **list= next(dpy, nextensions);
	if (!list)
		return list;

	for (int i = 0; i < *nextensions; ++i) {
		if (is_allowed(list[i]))
			continue;

		memmove(&list[i], &list[i + 1], (*nextensions - (i + 1)) * sizeof(*list));
		--*nextensions;
		list[*nextensions] = 0;
		--i;
	}

	return list;
}
